/* Quintuple Agent passphrase query module for X
 * Copyright (C) 1999 Robert Bihlmeyer <robbe@orcus.priv.at>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <assert.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <X11/X.h>


#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else 
#include "getopt.h"
#endif /* HAVE_GETOPT_H */

#include "assuan/assuan.h"
#include "gtksecentry.h"
#include "memory.h"
#include "util.h"

int debug = 0;

static int enhanced = 0;
static int global_grab = 1;

static GtkWidget *entry, *insure, *timeout;

static char *description_text;
static char *prompt_text;
static char *error_text;

static char *passphrase = NULL;
static int passphrase_ok = 0;

/* usage - show usage */
static void 
usage (void)
{
  fprintf(stderr, "Usage: gpinentry [OPTION]\n\
Ask securely for a secret and print it to stdout.\n\
Put QUESTION in the window, if given.\n\
\n\
  -e, --enhanced        ask for timeout and insurance, too\n\
  -g, --no-global-grab  grab keyboard only while window is focused\n\
  -d, --debug           turn on debugging output\n\
      --help            display this help and exit\n\
      --version         output version information and exit\n");
}

/* ok - print the passphrase to stdout and exit */
static void 
ok (GtkWidget *w, gpointer data)
{
  gtk_main_quit ();
}


/* unselect - work around a bug in GTK+ that permits word-selection to
   work on the invisible passphrase */
static void 
unselect(GtkWidget *w, GdkEventButton *e)
{
  static gint lastpos;

  if (e->button == 1) {
    if (e->type == GDK_BUTTON_PRESS) {
      lastpos = gtk_editable_get_position(GTK_EDITABLE(entry));
    } else if (e->type == GDK_2BUTTON_PRESS || e->type == GDK_3BUTTON_PRESS) {
      gtk_secure_entry_set_position(GTK_SECURE_ENTRY(w), lastpos);
      gtk_secure_entry_select_region(GTK_SECURE_ENTRY(w), 0, 0);
    }
  }
}

/* constrain_size - constrain size of the window the window should not
   shrink beyond the requisition, and should not grow vertically */
static void 
constrain_size(GtkWidget *win, GtkRequisition *req, gpointer data)
{
  static gint width, height;
  GdkGeometry geo;

  if (req->width == width && req->height == height)
    return; 
  width = req->width;
  height = req->height;
  geo.min_width = width;
  geo.max_width = 10000;	/* this limit is arbitrary,
				   but INT_MAX breaks other things */
  geo.min_height = geo.max_height = height;
  gtk_window_set_geometry_hints(GTK_WINDOW(win), NULL, &geo,
  				GDK_HINT_MIN_SIZE|GDK_HINT_MAX_SIZE);
}

/* grab_keyboard - grab the keyboard for maximum security */
static void 
grab_keyboard(GtkWidget *win, GdkEvent *event, gpointer data)
{
  if (gdk_keyboard_grab(win->window, FALSE, gdk_event_get_time(event))) 
    {
      g_error("could not grab keyboard");
    }
}

/* ungrab_keyboard - remove grab */
static void 
ungrab_keyboard(GtkWidget *win, GdkEvent *event, gpointer data)
{
  gdk_keyboard_ungrab(gdk_event_get_time(event));
}


static int
delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_main_quit ();
  return TRUE;
}

static void
button_clicked (GtkWidget *widget, gpointer data)
{
  if (data)
    { /* okay button */
      const char *s;
      
      fprintf (stderr, "OK button clicked\n");
      if (enhanced)
        {
          printf("Options: %s\nTimeout: %d\n\n",
                 gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(insure)) ?
                 "insure"
                 : "",
                 gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(timeout)));
        }

      s = gtk_secure_entry_get_text(GTK_SECURE_ENTRY(entry));
      if (!s)
        s = "";
      fprintf (stderr, "  got PIN `%s'\n", s);
      passphrase_ok = 1;
      secmem_free (passphrase);
      passphrase = secmem_malloc(strlen(s)+1);
      if (passphrase)
        strcpy (passphrase, s);
      fprintf (stderr, "  passphrase=`%s'\n", passphrase);
    }
  fprintf (stderr, "calling gtk_main_quit\n");
  gtk_main_quit ();
}


static GtkWidget *
create_window ()
{
  GtkWidget *w;
  GtkWidget *win, *box, *ebox;
  GtkWidget *sbox, *bbox;
  GtkAccelGroup *acc;
  int i;

  /* fixme: check the grabbing code against the one we used with the
     old gpg-agent */
  win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  acc = gtk_accel_group_new ();

  gtk_signal_connect (GTK_OBJECT (win), "delete_event",
                      GTK_SIGNAL_FUNC (delete_event), NULL);

  /*gtk_signal_connect (GTK_OBJECT(win), "destroy", gtk_main_quit, NULL);*/
  gtk_signal_connect (GTK_OBJECT(win), "size-request", constrain_size, NULL);
  gtk_signal_connect (GTK_OBJECT(win),
                      global_grab ? "map-event" : "focus-in-event",
                      grab_keyboard, NULL);
  gtk_signal_connect (GTK_OBJECT(win),
                      global_grab ? "unmap-event" : "focus-out-event",
                      ungrab_keyboard, NULL);

  gtk_accel_group_attach(acc, GTK_OBJECT(win));

  box = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER(win), box);

  if (description_text)
    {
      w = gtk_label_new (description_text);
      gtk_box_pack_start (GTK_BOX(box), w, FALSE, FALSE, 0);
    }
  if (error_text)
    {
      GtkStyle *style;
      GdkColormap *cmap;
      GdkColor color[1] = {{0, 0xffff, 0, 0}};
      gboolean success[1];

      w = gtk_label_new (error_text);
      gtk_box_pack_start (GTK_BOX(box), w, FALSE, FALSE, 0);

      /* fixme: Do we need to release something, or is tehre a more
         easy way to set a text color? */
      gtk_widget_realize (win);
      cmap = gdk_window_get_colormap (win->window);
      assert (cmap);
      gdk_colormap_alloc_colors (cmap, color, 1, FALSE, TRUE, success);
      if (success[0])
        {
          gtk_widget_ensure_style(w);
          style = gtk_style_copy(gtk_widget_get_style(w));
          style->fg[GTK_STATE_NORMAL] = color[0];
          gtk_widget_set_style(w, style);
        } 
    }

  ebox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX(box), ebox, FALSE, FALSE, 0);

  if (prompt_text)
    {
      w = gtk_label_new (prompt_text);
      gtk_box_pack_start (GTK_BOX(ebox), w, FALSE, FALSE, 0);
    }

  entry = gtk_secure_entry_new ();
  gtk_box_pack_start (GTK_BOX(ebox), entry, FALSE, FALSE, 0);
  gtk_secure_entry_set_visibility (GTK_SECURE_ENTRY(entry), FALSE);
  gtk_signal_connect_after (GTK_OBJECT(entry), "button_press_event",
                            unselect, NULL);
  gtk_widget_grab_focus (entry);
  gtk_widget_show (entry);

  if (enhanced)
    {
      sbox = gtk_hbox_new(FALSE, 5);
      gtk_box_pack_start(GTK_BOX(box), sbox, FALSE, FALSE, 0);
      
      w = gtk_label_new ("Forget secret after");
      gtk_box_pack_start(GTK_BOX(sbox), w, FALSE, FALSE, 0);
      gtk_widget_show(w);
	
      timeout = gtk_spin_button_new
        (GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, HUGE_VAL, 1, 60, 60)),2,0);
      gtk_box_pack_start (GTK_BOX(sbox), timeout, FALSE, FALSE, 0);
      gtk_widget_show (timeout);
	
      w = gtk_label_new ("seconds");
      gtk_box_pack_start (GTK_BOX(sbox), w, FALSE, FALSE, 0); 
      gtk_widget_show (w);
      gtk_widget_show (sbox);

      insure = gtk_check_button_new_with_label
        ("ask before giving out secret");
      gtk_box_pack_start (GTK_BOX(box), insure, FALSE, FALSE, 0);
      gtk_widget_show (insure);
    }
            
  bbox = gtk_hbutton_box_new();
  gtk_box_pack_start (GTK_BOX(box), bbox, FALSE, FALSE, 0);
  
  w = gtk_button_new_with_label ("OK");
  gtk_container_add (GTK_CONTAINER(bbox), w);
  gtk_signal_connect (GTK_OBJECT(w), "clicked", button_clicked, "ok");
  GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (w);
  gtk_signal_connect_object (GTK_OBJECT (entry), "focus_in_event",
    GTK_SIGNAL_FUNC (gtk_widget_grab_default), GTK_OBJECT (w));
  gtk_signal_connect_object (GTK_OBJECT (entry), "activate",
    GTK_SIGNAL_FUNC (gtk_button_clicked), GTK_OBJECT (w));

  gtk_widget_show (w);
  
  w = gtk_button_new_with_label ("Cancel");
  gtk_container_add (GTK_CONTAINER(bbox), w);
  gtk_accel_group_add (acc, GDK_Escape, 0, 0, GTK_OBJECT(w), "clicked");
  gtk_signal_connect (GTK_OBJECT(w), "clicked", button_clicked, NULL);
  GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);

  gtk_window_set_position (GTK_WINDOW (win), GTK_WIN_POS_CENTER);

  gtk_widget_show_all (win);

  return win;
}



static int
cmd_setdesc (ASSUAN_CONTEXT ctx, char *line)
{
  free (description_text);
  description_text = malloc (strlen(line)+1);
  if (!description_text)
    return ASSUAN_Out_Of_Core;
  strcpy (description_text, line);
  return 0;
}

static int
cmd_setprompt (ASSUAN_CONTEXT ctx, char *line)
{
  free (prompt_text);
  prompt_text = malloc (strlen(line)+1);
  if (!prompt_text)
    return ASSUAN_Out_Of_Core;
  strcpy (prompt_text, line);
  return 0;
}

static int
cmd_seterror (ASSUAN_CONTEXT ctx, char *line)
{
  free (error_text);
  error_text = malloc (strlen(line)+1);
  if (!error_text)
    return ASSUAN_Out_Of_Core;
  strcpy (error_text, line);
  return 0;
}

static int
cmd_getpin (ASSUAN_CONTEXT ctx, char *line)
{
  GtkWidget *w;
  int rc;

  passphrase_ok = 0;
  if (!prompt_text)
    prompt_text = strdup ("PIN:");
  if (!prompt_text)
    return ASSUAN_Out_Of_Core;
  w = create_window ();
  gtk_main();
  gtk_widget_destroy (w);
  while (gtk_events_pending ())
    gtk_main_iteration ();

  /* error text is not peristent - reset it now */
  free (error_text); error_text = NULL; 

  if (!passphrase_ok)
    return ASSUAN_Canceled;
  if (!passphrase)
    return ASSUAN_Out_Of_Core;
  rc = assuan_send_data (ctx, passphrase, strlen(passphrase));
  secmem_free (passphrase);
  passphrase = NULL;
  if (!rc)
    rc = assuan_send_data (ctx, NULL, 0);
  return rc;
}


static int
cmd_confirm (ASSUAN_CONTEXT ctx, char *line)
{
  return ASSUAN_Not_Implemented;
}


/* Tell the assuan library about our commands */
static int
register_commands (ASSUAN_CONTEXT ctx)
{
  static struct {
    const char *name;
    int cmd_id;
    int (*handler)(ASSUAN_CONTEXT, char *line);
  } table[] = {
    { "SETDESC",    0,  cmd_setdesc },
    { "SETPROMPT",  0,  cmd_setprompt },
    { "SETERROR",   0,  cmd_seterror },
    { "GETPIN",     0,  cmd_getpin },
    { "CONFIRM",    0,  cmd_confirm },
    { NULL }
  };
  int i, j, rc;

  for (i=j=0; table[i].name; i++)
    {
      rc = assuan_register_command (ctx,
                                    table[i].cmd_id? table[i].cmd_id
                                                   : (ASSUAN_CMD_USER + j++),
                                    table[i].name, table[i].handler);
      if (rc)
        return rc;
    } 
  return 0;
}


/* Startup the server */
static int
start_command_handler (void)
{
  int rc;
  int filedes[2];
  ASSUAN_CONTEXT ctx;

  /* For now we use a simple pipe based server so that we can work
     from scripts.  We will later add options to run as a daemon and
     wait for requests on a Unix domain socket */
  filedes[0] = 0;
  filedes[1] = 1;
  rc = assuan_init_pipe_server (&ctx, filedes);
  if (rc)
    {
      fprintf (stderr, "failed to initialize the server: %s\n",
               assuan_strerror(rc));
      return -1;
    }
  rc = register_commands (ctx);
  if (rc)
    {
      fprintf (stderr, "failed to the register commands with Assuan: %s\n",
               assuan_strerror(rc));
      return -1;
    }

  for (;;)
    {
      rc = assuan_accept (ctx);
      if (rc == -1)
          break;
      else if (rc)
        {
          fprintf (stderr, "Assuan accept problem: %s\n",
                   assuan_strerror (rc));
          break;
        }
      
      rc = assuan_process (ctx);
      if (rc)
        {
          fprintf (stderr, "Assuan processing failed: %s\n",
                   assuan_strerror (rc));
          continue;
        }
    }


  assuan_deinit_pipe_server (ctx);
  return 0;
}




int 
main(int argc, char **argv)
{
  int opt, opt_help = 0, opt_version = 0;
  struct option opts[] = {{ "debug",	      no_argument, NULL,	'd' },
			  { "enhanced",       no_argument, NULL,	'e' },
			  { "no-global-grab", no_argument, NULL,	'g' },
			  { "help",	      no_argument, &opt_help,	 1  },
			  { "version",	      no_argument, &opt_version, 1  },
			  { NULL, 0, NULL, 0 }};

  secmem_init(1);		/* 1 is too small, so default size is used */
  secmem_set_flags(SECMEM_WARN);
  drop_privs();

  /* fixme: initialize gtk_init only if DisPLAY is set, if it is not
     set, a curses based dialog should be used */
  gtk_init (&argc, &argv);
  while ((opt = getopt_long(argc, argv, "deg", opts, NULL)) != -1)
    {
      switch (opt)
        {
        case 'd':
          debug = 1;
          break;
        case 'e':
          enhanced = 1;
          break;
        case 'g':
          global_grab = 0;
          break;
        case 0:
        case '?':
          break;
        default:
          g_error("impossible case in %s:%d", __FILE__, __LINE__);
        }
    }
  if (opt_version) 
    {
      fprintf (stderr, "gpinentry " VERSION "\n");
      exit(EXIT_SUCCESS);
    }
  if (opt_help) 
    {
      usage();
      exit(EXIT_SUCCESS);
    }

  if (start_command_handler ())
    return 1;

  return 0;
}


